/*******************************************************************************
 * (c) Copyright 2011 Actel Corporation.  All rights reserved.
 *
 * G4Main Microcontroller Subsystem 10/100/1000 Mbps Ethernet MAC metal software
 * driver implementation.
 *
 * SVN $Revision: 4969 $
 * SVN $Date: 2013-01-11 21:50:58 +0000 (Fri, 11 Jan 2013) $
 */
#include <stdio.h>
#include "../../CMSIS/m2sxxx.h"
//<CJ>#include "../../CMSIS/core_cm3.h"
#include "../../CMSIS/mss_assert.h"
#include "../../CMSIS/hw_reg_io.h"
#include "mss_ethernet_mac_regs.h"
#include "mss_ethernet_mac.h"
#include "phy.h"

#ifdef __cplusplus
extern "C" {
#endif

/**************************************************************************/
/* Preprocessor Macros                                                    */
/**************************************************************************/
#define SET       ENABLE
#define RESET     DISABLE

#define IS_STATE(x)             ( ((x) == ENABLE) || ((x) == DISABLE) )
#define IS_ENABLE(x)            ( (x) == ENABLE )
#define IS_DISABLE(x)           ( (x) == DISABLE )

#define IS_MAC_MODE(x)          ( ((x) == MSS_MAC_NIBBLE_MODE ) || \
                                  ((x) == MSS_MAC_BYTE_MODE ) )
#define IS_BYTE_MODE(x)         ( ((x) == MSS_MAC_BYTE_MODE ) )
#define IS_NIBBLE_MODE(x)       ( ((x) == MSS_MAC_NIBBLE_MODE) )

#define GET_DESC_EMPTY_FLAG(x)          ( (x) & DMA_DESC_EMPTY_FLAG_MASK )
#define SET_DESC_EMPTY_FLAG(x)          ( (x) = (x) | (DMA_DESC_EMPTY_FLAG_MASK) )
#define CLEAR_DESC_EMPTY_FLAG(x)        ( (x) = (x) & (uint32_t)(~DMA_DESC_EMPTY_FLAG_MASK) )
#define GET_PKT_SIZE(x)                 ( (x) & DMA_DESC_PKT_SIZE_MASK )

#define MAC_STAT_REG_OFFSET             ( 0x00000080u )
#define MAC_STAT_BASE                   ( ETHERNET_BASE + MAC_STAT_REG_OFFSET )
#define MAC_STAT_REG(x)                 ( (uint32_t *)( MAC_STAT_BASE + ((x) << 2u)) )

#define IS_WORD_ALIGNED(x)      ((uint32_t)0 == ((uint32_t)x & (uint32_t)3))

/**************************************************************************/
/* Type defines                                                           */
/**************************************************************************/

/**************************************************************************/
/* Global variables                                                       */
/**************************************************************************/

MAC_instance_t g_mac;

/**************************************************************************/
/* Private Functions                                                      */
/**************************************************************************/
static void MAC_desc_ring_init(MAC_desc_t * ring, MAC_io_t io);
static uint8_t * MAC_memcpy(uint8_t *dest, const uint8_t *src, uint32_t n);
static return_t MAC_assign_station_addr(const uint8_t *mac_addr);
static void MAC_rxpkt_handler(void);
static void MAC_txpkt_handler(void);
static return_t MAC_auto_link_setup( MAC_cfg_t * cfg );

/**************************************************************************/
/* Public Functions                                                       */
/**************************************************************************/

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_init
( 
    MAC_cfg_t * cfg 
)
{
    return_t status = FAILED;
    const uint8_t def_mac_address[6] = { MSS_MAC_DEFAULT_MAC_ADDRESS };
    uint32_t tempreg = 0;

    ASSERT( cfg != NULL );

    if( cfg != NULL )
    {
        /* Check for validity of configuration parameters */
        ASSERT( INVALID_INF > cfg->interface );
        ASSERT( IS_STATE(cfg->tx_edc_enable) );
        ASSERT( IS_STATE(cfg->rx_edc_enable) );        
        ASSERT( IS_STATE(cfg->auto_negotiation) ); 
        ASSERT( MSS_MAC_PREAMLEN_MAXVAL >= cfg->preamble_length );
        ASSERT( IS_MAC_MODE(cfg->interface_mode) );
        ASSERT( IS_STATE(cfg->hugeframe_enable) );
        ASSERT( IS_STATE(cfg->length_field_check) );
        ASSERT( IS_STATE(cfg->pad_n_CRC) );
        ASSERT( IS_STATE(cfg->append_CRC) );
        ASSERT( IS_STATE(cfg->loopback) );
        ASSERT( IS_STATE(cfg->rx_flow_ctrl) );
        ASSERT( IS_STATE(cfg->tx_flow_ctrl) );
        ASSERT( MSS_MAC_MINIFG_MAXVAL >= cfg->min_IFG );            
        ASSERT( MSS_MAC_BTBIFG_MAXVAL >= cfg->btb_IFG );           
        ASSERT( MSS_MAC_MAXRETX_MAXVAL >= cfg->max_retx_tries );      
        ASSERT( IS_STATE(cfg->excessive_defer) );    
        ASSERT( IS_STATE(cfg->nobackoff) );         
        ASSERT( IS_STATE(cfg->backpres_nobackoff) );
        ASSERT( IS_STATE(cfg->ABEB_enable) );        
        ASSERT( MSS_MAC_ABEBTRUNC_MAXVAL >= cfg->ABEB_truncvalue );    
        ASSERT( MSS_MAC_BY28_PHY_CLK >= cfg->phyclk );            
        ASSERT( IS_STATE(cfg->supress_preamble) );   
        ASSERT( IS_STATE(cfg->autoscan_phys) );      
        ASSERT( LAST_PHY > cfg->phy );           
        ASSERT( MSS_MAC_MAXFRAMELEN_MAXVAL >= cfg->max_frame_length );   
        ASSERT( MSS_MAC_SLOTTIME_MAXVAL >= cfg->slottime );   

        /* 10/100/1000 Mbps speed configuration for MUXes */
        /* Configure Interface speed */
        if( MAC100MBPS == cfg->speed )
            set_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
        else
            clear_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
        tempreg = cfg->speed;
        /* Interface type: MII/GMII/RMII/RGMII/SGMII */
        tempreg |= ( ((uint32_t)cfg->interface) << MAC_CONFIG_INTF );
        SYSREG->MAC_CR = tempreg;

        /* Tx FIFO EDAC configuration */
        ( IS_ENABLE(cfg->tx_edc_enable) ? set_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_TX_EN) :
                                          clear_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_TX_EN) );
        /* Rx FIFO EDAC configuration */
        ( IS_ENABLE(cfg->rx_edc_enable) ? set_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_RX_EN) :
                                          clear_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_RX_EN) );

        /*********************************************************************/
        /* Auto detection of PHY if phy is configured for AUTO_PHY           */
        /* Not yet supported                                                 */
        /*********************************************************************/

        /* Configure PHY related MII MGMT registers */
        tempreg = (uint32_t)cfg->phyclk;
        ( IS_ENABLE(cfg->supress_preamble) ? (tempreg |= MII_PREAM_SUPRESS_MASK):
                                             (tempreg &= ~MII_PREAM_SUPRESS_MASK) );
        ( IS_ENABLE(cfg->autoscan_phys) ? (tempreg |= MII_SCAN_AUTO_INC_MASK):
                                             (tempreg &= ~MII_SCAN_AUTO_INC_MASK) );  
        MAC->MII_CONFIG = tempreg;

        tempreg = ((uint32_t)(PHY_device[cfg->phy].addr)) << MII_PHY_ADDR;
        MAC->MII_ADDRESS |= tempreg;

        /* Initialize MAC instance */
        g_mac.hw_reg = MAC;
        g_mac.irqn   = EthernetMAC_IRQn;
        g_mac.cfg    = cfg;
        g_mac.phy    = cfg->phy;
        g_mac.stat   = MAC_LAST_STAT;

        /* Intialize Tx & Rx descriptor rings */
        MAC_desc_ring_init( &g_mac.tx_desc_tab[0], TX);
        MAC_desc_ring_init( &g_mac.rx_desc_tab[0], RX);
        
        /* 
         * At the time of intialization next descriptors
         * and Current descriptors will be NULL as no tx/rx is performed.
         */
        g_mac.curr_tx_desc = NULL;
        g_mac.next_tx_desc = &g_mac.tx_desc_tab[0];
        g_mac.curr_rx_desc = NULL;
        g_mac.next_rx_desc = &g_mac.rx_desc_tab[0];   
        /*
         * MAC hardware configuration
         */              
        /* Clear soft reset to keep MCXMAC functional */
        clear_bit_reg32( &MAC->CFG1, CFG1_SOFT_RST);

        /* Clear all reset bits */
        /* Clear MCXMAC TX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_TX_RST);
        /* Clear MCXMAC RX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_RX_RST);
        /* Clear MCXMAC TX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_TXCTL_RST);
        /* Clear MCXMAC RX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_RXCTL_RST);
        /* Clear MCXMAC interface reset */
        clear_bit_reg32( &MAC->INTERFACE_CTRL, MII_INTF_RESET);
        /* Clear FIFO watermark reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_WMM_RST);
        /* Clear FIFO Rx system reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RSYS_RST);
        /* Clear FIFO Rx fab reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RFAB_RST);
        /* Clear FIFO Tx system reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TSYS_RST);
        /* Clear FIFO Tx system reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TFAB_RST);

        ( IS_ENABLE(cfg->loopback) ? ( tempreg = CFG1_LOOPBACK_MASK) : (tempreg = 0) );
        ( IS_ENABLE(cfg->rx_flow_ctrl) ? ( tempreg |= CFG1_RX_FCTL_MASK) :
                                         ( tempreg &= ~CFG1_RX_FCTL_MASK));                                     
        ( IS_ENABLE(cfg->tx_flow_ctrl) ? ( tempreg |= CFG1_TX_FCTL_MASK) :
                                         ( tempreg &= ~CFG1_TX_FCTL_MASK));
        MAC->CFG1 = tempreg ;

        ( IS_ENABLE(cfg->fullduplex) ? ( tempreg = CFG2_FDX_MASK) : (tempreg = 0) );
        ( IS_ENABLE(cfg->append_CRC) ? ( tempreg |= CFG2_CRC_EN_MASK) :
                                       ( tempreg &= ~CFG2_CRC_EN_MASK));
        ( IS_ENABLE(cfg->pad_n_CRC) ? ( tempreg |= CFG2_PAD_CRC_EN_MASK) :
                                      ( tempreg &= ~CFG2_PAD_CRC_EN_MASK));
        ( IS_ENABLE(cfg->length_field_check) ? ( tempreg |= CFG2_LEN_CHECK_MASK) :
                                               ( tempreg &= ~CFG2_LEN_CHECK_MASK));        
        ( IS_ENABLE(cfg->hugeframe_enable) ? ( tempreg |= CFG2_HUGE_FRAME_EN_MASK) :
                                             ( tempreg &= ~CFG2_HUGE_FRAME_EN_MASK));
        if( IS_NIBBLE_MODE( cfg->interface_mode ) )
        {
            tempreg |= CFG2_NIBBLE_MASK;
        }
        else
        {
            tempreg |= CFG2_BYTE_MASK;
        }
        tempreg |= ( ((uint32_t)cfg->preamble_length ) << CFG2_PREAM_LEN );
        MAC->CFG2 = tempreg ;        
        
        tempreg = cfg->btb_IFG;
        tempreg |= ( (uint32_t)cfg->min_IFG << IFG_MINIFGENF );
        tempreg |= ( cfg->non_btb_IFG << IFG_NONBTBIPG );
        MAC->IFG = tempreg;
        
        tempreg = cfg->slottime;
        tempreg |= ( (uint32_t)cfg->max_retx_tries << HALF_DUPLEX_RETX_MAX);
        ( IS_ENABLE(cfg->excessive_defer)? (tempreg |= HALF_DUPLEX_EXCS_DEFER_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_EXCS_DEFER_MASK) ); 
        ( IS_ENABLE(cfg->nobackoff)? (tempreg |= HALF_DUPLEX_NO_BACKOFF_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_NO_BACKOFF_MASK) );
        ( IS_ENABLE(cfg->backpres_nobackoff)? (tempreg |= HALF_DUPLEX_BACKPRES_NOBACKOFF_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_BACKPRES_NOBACKOFF_MASK) );
        ( IS_ENABLE(cfg->ABEB_enable)? (tempreg |= HALF_DUPLEX_ABEB_ENABLE_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_ABEB_ENABLE_MASK) );       
        tempreg |= ( (uint32_t)cfg->ABEB_truncvalue << HALF_DUPLEX_ABEB_TUNC );
        MAC->HALF_DUPLEX = tempreg;

        MAC->MAX_FRAME_LENGTH = (uint32_t)cfg->max_frame_length;

        /* Configure FIFO */
        MAC->FIFO_CFG0 = ( FIFO_CFG0_WMM_ENABLE  | FIFO_CFG0_RSYS_ENABLE |
                           FIFO_CFG0_RFAB_ENABLE | FIFO_CFG0_TSYS_ENABLE |
                           FIFO_CFG0_TFAB_ENABLE );
        /* RX FIFO size : 8KB  */        
        MAC->FIFO_CFG1 = FIFO_CFG1_DEFVAL;
        /* Rx FIFO watermark */
        MAC->FIFO_CFG2 = FIFO_CFG2_DEFVAL;
        /* Tx FIFO watermark: 4KB Tx FIFO */
        MAC->FIFO_CFG3 = FIFO_CFG3_DEFVAL;
        /* Drop junk frames (Frames suffered CRC error or short collision) */
        MAC->FIFO_CFG4 = FIFO_CFG4_DEFVAL;
        /* Drop frames with CRC errors and PAUSE frames */
        MAC->FIFO_CFG5 = FIFO_CFG5_DEFVAL;

        /* Assgin default MAC station address */
        status = MAC_assign_station_addr( def_mac_address );

        /* initialize default interrupt handlers */
        g_mac.callback[MSS_MAC_TXPKTSENT_IRQ]   = NULL;
        g_mac.callback[MSS_MAC_TXUNDRRUN_IRQ]   = NULL;
        g_mac.callback[MSS_MAC_TXUNDRRUN_IRQ+1] = NULL;
        g_mac.callback[MSS_MAC_TXBUSERR_IRQ]    = NULL;
        g_mac.callback[MSS_MAC_RXPKTRCVD_IRQ]   = NULL;
        g_mac.callback[MSS_MAC_RXPKTRCVD_IRQ+1] = NULL;
        g_mac.callback[MSS_MAC_RXOVRFLOW_IRQ]   = NULL;
        g_mac.callback[MSS_MAC_RXBUSERR_IRQ]    = NULL;

        /* Enable transmission at MAC level. */
         set_bit_reg32( &MAC->CFG1, CFG1_TX_EN);
        /* Enable reception at MAC level.    */
         set_bit_reg32( &MAC->CFG1, CFG1_RX_EN);

        /* Initialize PHY interface */
        PHY_device[cfg->phy].init( cfg );

        /* Auto Negotiate link parameters */
        /* This modifies the cfg to the negotiated parameters */
        /* And the MAC will be initialized with this cfg */
        if( ( MSS_MAC_AUTO_NEGOTIATION_ENABLE == cfg->auto_negotiation ) ||
          ( TBI == cfg->interface ) )
        {
            status = MAC_auto_link_setup( cfg );
            if( SUCCESS != status )
                return status;          /* Failing here */
        }
        /* Force setup only for other interfaces */
        else
        {
            /* Configure speed and duplex mode */
            phy_setup_forced(cfg->phy, cfg);
        }
        /* Check for correctness of negotiated parameters */
        ASSERT( INVALID_SPEED > cfg->speed );                    
        ASSERT( IS_STATE(cfg->fullduplex) );
        
        /* Reconfigure for auto-negotiated parameters. */
        /* 10/100/1000 Mbps speed configuration for MUXes */
        /* Configure Interface speed */
        if( MAC100MBPS == cfg->speed )
            set_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
        else
            clear_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
        tempreg = cfg->speed;
        /* Interface type: MII/GMII/RMII/RGMII/SGMII */
        tempreg |= ( ((uint32_t)cfg->interface) << MAC_CONFIG_INTF );
        SYSREG->MAC_CR = tempreg;
        /* Configure duplex mode */
        ( IS_ENABLE(cfg->fullduplex) ? set_bit_reg32(&MAC->CFG2,CFG2_FDX) : 
                                    clear_bit_reg32(&MAC->CFG2,CFG2_FDX) );

        /* Enable DMA Transmission and Reception is out of scope
           of this function. You need enable the transmission, reception
           and corresponding interrupts, if required as part of the
           system initialization. */

        status = status && SUCCESS;
    }

    return status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_cfg_struct_def_init
(
    MAC_cfg_t * cfg
)
{
    return_t status = FAILED;
    ASSERT( NULL != cfg);
    if( NULL != cfg )
    {
        cfg->speed              = MAC10MBPS;	               
        cfg->interface          = RMII;
        cfg->auto_negotiation   = MSS_MAC_AUTO_NEGOTIATION_ENABLE;      
        cfg->tx_edc_enable      = MSS_MAC_ERR_DET_CORR_ENABLE;            
        cfg->rx_edc_enable      = MSS_MAC_ERR_DET_CORR_ENABLE;            
        cfg->preamble_length    = MSS_MAC_PREAMLEN_DEFVAL;       
        cfg->interface_mode     = MSS_MAC_NIBBLE_MODE;         
        cfg->hugeframe_enable   = MSS_MAC_HUGE_FRAME_DISABLE;      
        cfg->length_field_check = MSS_MAC_LENGTH_FILED_CHECK_ENABLE;    
        cfg->pad_n_CRC          = MSS_MAC_PAD_N_CRC_ENABLE;			   
        cfg->append_CRC         = MSS_MAC_CRC_ENABLE;            
        cfg->fullduplex         = MSS_MAC_FULLDUPLEX_ENABLE;            
        cfg->loopback           = MSS_MAC_LOOPBACK_DISABLE;              
        cfg->rx_flow_ctrl       = MSS_MAC_RX_FLOW_CTRL_DISABLE;          
        cfg->tx_flow_ctrl       = MSS_MAC_TX_FLOW_CTRL_DISABLE;          
        cfg->min_IFG            = MSS_MAC_MINIFG_DEFVAL;               
        cfg->btb_IFG            = MSS_MAC_BTBIFG_DEFVAL;               
        cfg->max_retx_tries     = MSS_MAC_MAXRETX_DEFVAL;        
        cfg->excessive_defer    = MSS_MAC_EXSS_DEFER_DISABLE;       
        cfg->nobackoff          = MSS_MAC_NO_BACKOFF_DISABLE;             
        cfg->backpres_nobackoff = MSS_MAC_BACKPRESS_NO_BACKOFF_DISABLE;    
        cfg->ABEB_enable        = MSS_MAC_ABEB_DISABLE;           
        cfg->ABEB_truncvalue    = MSS_MAC_ABEBTRUNC_DEFVAL;       
        cfg->phyclk             = MSS_MAC_DEF_PHY_CLK;                
        cfg->supress_preamble   = MSS_MAC_SUPPRESS_PREAMBLE_DISABLE;      
        cfg->autoscan_phys      = MSS_MAC_PHY_AUTOSCAN_DISABLE;         
        cfg->phy                = MSS_MAC_DEFAULT_PHY;               
        cfg->max_frame_length   = MSS_MAC_MAXFRAMELEN_DEFVAL;     
        cfg->non_btb_IFG        = MSS_MAC_NONBTBIFG_DEFVAL;          
        cfg->slottime           = MSS_MAC_SLOTTIME_DEFVAL;             
        cfg->framedrop_mask     = MSS_MAC_FRAME_DROP_MASK_DEFVAL;       

        status = SUCCESS;
    }                    
    return status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t 
MSS_MAC_read_tx_status
(
    MAC_status_t status_bit
)
{
    return (read_bit_reg32( &MAC->DMA_TX_STATUS, status_bit));
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t 
MSS_MAC_read_rx_status
(
    MAC_status_t status_bit
)
{
    return (read_bit_reg32( &MAC->DMA_RX_STATUS, status_bit));
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t 
MSS_MAC_read_tx_pkt_count
(
    void
)
{
    return ( (uint8_t)( (MAC->DMA_TX_STATUS & DMA_PKTCOUNT_MASK) >> 
                                                     DMA_PKTCOUNT) );
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t 
MSS_MAC_read_rx_pkt_count
(
    void
)
{
    return ( (uint8_t)( (MAC->DMA_RX_STATUS & DMA_PKTCOUNT_MASK) >> 
                                                     DMA_PKTCOUNT) );
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void 
MSS_MAC_reset
(
    void
)
{
    /* Reset MCXMAC TX functionality */
    set_bit_reg32( &MAC->CFG1, CFG1_TX_RST);
    /* Reset MCXMAC RX functionality */
    set_bit_reg32( &MAC->CFG1, CFG1_RX_RST);
    /* Reset MCXMAC TX control */
    set_bit_reg32( &MAC->CFG1, CFG1_TXCTL_RST);
    /* Reset MCXMAC RX control */
    set_bit_reg32( &MAC->CFG1, CFG1_RXCTL_RST);
    /* Reset MCXMAC interface control */
    set_bit_reg32( &MAC->INTERFACE_CTRL, MII_INTF_RESET);
    /* Reset FIFO watermark module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_WMM_RST);
    /* Reset FIFO Rx system module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RSYS_RST);
    /* Reset FIFO Rx fab module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RFAB_RST);
    /* Reset FIFO Tx system module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TSYS_RST);
    /* Reset FIFO Tx system module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TFAB_RST);

    /* Intialize Tx & Rx descriptor rings */
    MAC_desc_ring_init( &g_mac.tx_desc_tab[0], TX);
    MAC_desc_ring_init( &g_mac.rx_desc_tab[0], RX);
    
    /* 
     * At the time of intialization next descriptors
     * and Current descriptors will be NULL as no tx/rx is performed.
     */
    g_mac.curr_tx_desc = NULL;
    g_mac.next_tx_desc = &g_mac.tx_desc_tab[0];
    g_mac.curr_rx_desc = NULL;
    g_mac.next_rx_desc = &g_mac.rx_desc_tab[0];
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t
MSS_MAC_write_phy_reg
(
    uint8_t phyaddr,
    uint8_t regaddr,
    uint16_t regval
)
{
    return_t status = FAILED;

    ASSERT( MSS_MAC_PHYADDR_MAXVAL >= phyaddr); 
    ASSERT( MSS_MAC_PHYREGADDR_MAXVAL >= regaddr );
    /* 
     * Write PHY address in MII Mgmt address register.
     * Makes previous register address 0 & invalid.
     */ 
    if((MSS_MAC_PHYADDR_MAXVAL >= phyaddr) && 
       (MSS_MAC_PHYREGADDR_MAXVAL >= regaddr))
    {
        /* Wait for MII Mgmt interface to complete previous operation. */
        do {
            ;
        } while(MAC->MII_INDICATORS & MII_BUSY_INDICATOR_MASK);
    
        /* Load PHY address in MII Mgmt address register */
        MAC->MII_ADDRESS = (((uint32_t)phyaddr)<<MII_PHY_ADDR);
        /* Load register address in MII Mgmt address register */
        MAC->MII_ADDRESS |= (uint32_t)regaddr;
        /* Load value to be written in MII Mgmt control register */
        MAC->MII_CTRL = (uint32_t)regval ;
        status = SUCCESS;
    }

    return status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint16_t
MSS_MAC_read_phy_reg
(
    uint8_t phyaddr,
    uint8_t regaddr
)
{
    uint32_t timeout = 100000u;
    uint16_t read_val = 0;

    ASSERT( MSS_MAC_PHYADDR_MAXVAL >= phyaddr); 
    ASSERT( MSS_MAC_PHYREGADDR_MAXVAL >= regaddr );
    /* 
     * Write PHY address in MII Mgmt address register.
     * Makes previous register address 0 & invalid.
     */ 
    if( ( MSS_MAC_PHYADDR_MAXVAL >= phyaddr) && 
        ( MSS_MAC_PHYREGADDR_MAXVAL >= regaddr ) )
    {
        /* Wait for MII Mgmt interface to complete previous operation. */
        do {
            ;
        } while(MAC->MII_INDICATORS & MII_BUSY_INDICATOR_MASK);
        
        MAC->MII_ADDRESS = (((uint32_t)phyaddr) << MII_PHY_ADDR);
        /* Load PHY register address in MII Mgmt address register */
        MAC->MII_ADDRESS |= (uint32_t)regaddr;
        /* Issue read command to PHY */
        set_bit_reg32( &MAC->MII_COMMAND, MII_READ);
        /* Poll till PHY read cycle is completed or timeout expired */
        while ( timeout-- && ((MAC->MII_INDICATORS & MII_BUSY_INDICATOR_MASK) || 
                              (MAC->MII_INDICATORS & MII_NV_INDICATOR_MASK)) );
        /* Read value from MII Mgmt status register only if timeout is not expired
           but the read the read cycle is completed */
        if( timeout != 0 )
            read_val = (uint16_t)MAC->MII_STATUS ;
    }
    clear_bit_reg32( &MAC->MII_COMMAND, MII_READ);
    return read_val;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_set_mac_addr
(
    uint8_t *address
)
{
    return_t status = FAILED;

    ASSERT( NULL!= address );            
    if( NULL != address )
    {
        MAC_memcpy(g_mac.mac_addr, address, 6);
        status = MAC_assign_station_addr(g_mac.mac_addr);
        status = status && SUCCESS;
    }

    return status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint32_t 
MSS_MAC_read_stat
(
    MAC_stat_t stat
)
{
    uint32_t stat_val = 0;

    ASSERT ( MAC_LAST_STAT > stat );
    if( MAC_LAST_STAT > stat )
        stat_val = *(MAC_STAT_REG(stat));

    return stat_val;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t
MSS_MAC_read_interface_status
(
    MAC_intf_status_t intf_status_bit    
)
{
    uint8_t inf_status = RESET;

    ASSERT( INVALID_INTF > intf_status_bit );
    if( INVALID_INTF > intf_status_bit )
        if( read_bit_reg32( (volatile uint32_t *)&MAC->INTERFACE_STATUS,
                            intf_status_bit ) )
            inf_status = SET;

    return inf_status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t
MSS_MAC_read_phy_indicator
(
    MAC_phy_ind_t indicator
)
{
    uint8_t ind_status = RESET;

    ASSERT( LAST_INDICATOR > indicator );
    if( LAST_INDICATOR > indicator )
        if( read_bit_reg32( (volatile uint32_t *)&MAC->MII_INDICATORS, 
                             indicator) )
            ind_status = SET;
    return ind_status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_init_pkt_pool
(
    MAC_pkt_t * pktring,
    MAC_io_t io
)
{
    return_t status = FAILED;
    uint32_t i;
    uint32_t len;
    ASSERT( NULL != pktring );

    if( NULL != pktring )
    {
        ( (TX == io) ? (len = MSS_MAC_TX_RING_SIZE) : 
                       (len = MSS_MAC_TX_RING_SIZE) );
        
        for(i=0; i<len; i++)
        {
            pktring[i].lock = EMPTY;
            pktring[i].buff = NULL;
            pktring[i].size = 0u;
        }
        status = SUCCESS;
    }
    return status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_is_pkt_received
(
    MAC_pkt_t * pkt 
)
{
    return ( UNLOCK == pkt->lock );
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_is_pkt_transmitted
(
    MAC_pkt_t * pkt 
)
{
    return ( UNLOCK == pkt->lock );
}


/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_transfer_rx_pkt
(
    uint8_t * dest,
    MAC_pkt_t * pkt    
)
{
    return_t status = FAILED;
    ASSERT( NULL != dest );
    ASSERT( NULL != pkt );

    if( ( NULL != dest ) && ( NULL != pkt ) )
    {
        if( dest != pkt->buff )
            MAC_memcpy( dest, pkt->buff, pkt->size );
        pkt->lock = EMPTY;
        status = SUCCESS;
    }
    return status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_prepare_tx_pkt
(
    uint8_t * source,
    MAC_pkt_t * pkt    
)
{
    return_t status = FAILED;
    ASSERT( NULL != source );
    ASSERT( NULL != pkt );

    if( ( NULL != source ) && ( NULL != pkt ) )
    {
        if( pkt->buff != source )
              MAC_memcpy( pkt->buff, source, pkt->size );
        pkt->lock = FULL;
        status = SUCCESS;
    }
    return status;
}
/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
MAC_pkt_t *
MSS_MAC_alloc_pkt
(
    MAC_pkt_t *pktring,
    MAC_io_t  io
)
{
    uint8_t state;
    uint32_t i;
    uint32_t len;
    MAC_pkt_t * pkt = NULL;
    ASSERT( NULL != pktring );

    if( NULL != pktring )
    {
        ( (TX == io) ? (len = MSS_MAC_TX_RING_SIZE, state = UNLOCK ) : 
                       (len = MSS_MAC_TX_RING_SIZE, state = EMPTY) );
        
        for(i=0; i<len; i++)
        {
            if( state == pktring[i].lock )
            {
                pkt = &pktring[i];
                break;
            }
        }
    }
    return pkt;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t
MSS_MAC_receive_pkt
(
    MAC_pkt_t * pkt,
    uint8_t  slots
)
{
    return_t status = FAILED;

    /* Make this function atomic w.r.to EMAC interrupt */
    NVIC_DisableIRQ(EthernetMAC_IRQn);

    ASSERT( NULL != pkt );
    ASSERT( IS_WORD_ALIGNED(pkt->buff) );
    ASSERT( EMPTY == pkt->lock );

    if((0u < slots) && (slots <= MSS_MAC_RX_RING_SIZE))
    {     
        while(slots)
        {
            if( ( NULL != pkt ) && ( EMPTY == pkt->lock ) && 
                                   ( IS_WORD_ALIGNED(pkt->buff) ) )  
            {
                if( ( NULL != g_mac.next_rx_desc ) && 
                    ( !GET_DESC_EMPTY_FLAG(g_mac.next_rx_desc->pkt_size) ) )    
                {
                    pkt->lock = LOCK;
                    g_mac.next_rx_desc->pkt_start_addr = (uint32_t)pkt->buff;
                    /* DMA writes received no.of bytes here */
                    g_mac.next_rx_desc->pkt_size |= 0;
                    g_mac.next_rx_desc->pkt = pkt;
        
                    /* Set own flag of this descriptor to indicate ready for RX */
                    SET_DESC_EMPTY_FLAG(g_mac.next_rx_desc->pkt_size);
        
                    /* 
                       If the RX is found disabled, then it might be because this is the
                       first time a packet is scheduled for reception or the RX ENABLE is 
                       made zero by RX overflow or RX bus error. In either case, this
                       function tries to schedule the current packet for reception.
                    */
                    if( !read_bit_reg32(&MAC->DMA_RX_CTRL, DMA_RXENABLE) )
                    {
                        /* Make current descriptor to point to the packet requested */
                        g_mac.curr_rx_desc = g_mac.next_rx_desc;
                        MAC->DMA_RX_DESC = (uint32_t)g_mac.curr_rx_desc;
                        set_bit_reg32( &MAC->DMA_RX_STATUS, DMA_RXOVRFLOW );
                        set_bit_reg32( &MAC->DMA_RX_CTRL, DMA_RXENABLE );
                    }
    
                    /* Point the next_rx_desc to next free descriptor in the ring */
                    /* Wrap around in case next descriptor is pointing to last in the ring */
                    if( (MSS_MAC_RX_RING_SIZE - 1u) == g_mac.next_rx_desc->index )
                        g_mac.next_rx_desc = &g_mac.rx_desc_tab[0];
                    else
                        ++g_mac.next_rx_desc;
                        
                    /* Packet scheduled for reception successfully only when there 
                       is no new RX bus error */
                    if( !read_bit_reg32( &MAC->DMA_RX_STATUS, DMA_RXBUSERR ) )
                        status = SUCCESS;
                }
            }
            /* Next packet if any */
            if( 0u != --slots )
                pkt++;
        }
    }
    /* Ethernet Interrupt Enable function. */
    NVIC_EnableIRQ(EthernetMAC_IRQn);
    return status;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_send_pkt
(
    MAC_pkt_t * pkt,
    uint8_t  slots
)
{
    return_t status = FAILED;

    /* Make this function atomic w.r.to EMAC interrupt */
    NVIC_DisableIRQ(EthernetMAC_IRQn);

    ASSERT( NULL != pkt );
    ASSERT( IS_WORD_ALIGNED(pkt->buff) );
    ASSERT( 0 != pkt->size );
    ASSERT( FULL == pkt->lock );

    if((0u < slots) && (slots <= MSS_MAC_TX_RING_SIZE))
    {
        while(slots)
        {
            if( (NULL != pkt) && (0 != pkt->size) && 
                ( FULL == pkt->lock) && ( IS_WORD_ALIGNED(pkt->buff) ) )
            {
                if( ( NULL != g_mac.next_tx_desc ) && 
                    ( GET_DESC_EMPTY_FLAG(g_mac.next_tx_desc->pkt_size) ) )
                {
                    pkt->lock = LOCK;
                    g_mac.next_tx_desc->pkt_start_addr = (uint32_t)pkt->buff;
                    g_mac.next_tx_desc->pkt_size |= pkt->size;
                    g_mac.next_tx_desc->pkt = pkt;
                    
                    /* Make own flag 0 for this descriptor to indicate ready for tx */
                    CLEAR_DESC_EMPTY_FLAG(g_mac.next_tx_desc->pkt_size);
        
                    /*
                        If TX is found disabled, this might be because this is
                        the first time packet is being sent or the DMA completed 
                        previous transfer and stopped transmission by itself caused
                        by TX underrun or bus error. This function neglects the errors
                        and tries to send the current packet.
                     */
                    if( !read_bit_reg32(&MAC->DMA_TX_CTRL, DMA_TXENABLE) )
                    {
                        g_mac.curr_tx_desc = g_mac.next_tx_desc;
                        MAC->DMA_TX_DESC = (uint32_t)g_mac.curr_tx_desc;
                        set_bit_reg32( &MAC->DMA_TX_STATUS, DMA_TXUNDRRUN );
                        set_bit_reg32( &MAC->DMA_TX_CTRL, DMA_TXENABLE );
                    }
        
                    /* Point the next_tx_desc to next free descriptor in the ring */
                    /* Wrap around in case next descriptor is pointing to last in the ring */
                    if( (MSS_MAC_TX_RING_SIZE - 1u) == g_mac.next_tx_desc->index )
                        g_mac.next_tx_desc = &g_mac.tx_desc_tab[0];
                    else
                        ++g_mac.next_tx_desc;
                    /* Packet scheduled for transmission successfully only when 
                       there is no TX bus error */
                    if( !read_bit_reg32( &MAC->DMA_TX_STATUS, DMA_TXBUSERR) )
                        status = SUCCESS;
                }
            }
        /* Next packet if any */
            if( 0u != --slots )
                pkt++;
        }
    }
    /* Ethernet Interrupt Enable function. */
    NVIC_EnableIRQ(EthernetMAC_IRQn);
    return status;
}
/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void
EthernetMAC_IRQHandler
(
    void
)
{
    /* Transmit packet sent interrupt */
    if((MAC->DMA_IRQ & MAC->DMA_IRQ_MASK) & (1u<<MSS_MAC_TXPKTSENT_IRQ))
    {
        /* Top-half */
        MAC_txpkt_handler();
    }
    /* Transmitter underrun interrupt */
    if((MAC->DMA_IRQ & MAC->DMA_IRQ_MASK) & (1u<<MSS_MAC_TXUNDRRUN_IRQ))
    {
        /* Call the user's custom handler */
        if( NULL != g_mac.callback[MSS_MAC_TXPKTSENT_IRQ])
                g_mac.callback[MSS_MAC_TXPKTSENT_IRQ]();
    }
    /* Transmitter bus error interrupt */
    if((MAC->DMA_IRQ & MAC->DMA_IRQ_MASK) & (1u<<MSS_MAC_TXBUSERR_IRQ))
    {
        /* Call the user's custom handler */
        if( NULL != g_mac.callback[MSS_MAC_TXPKTSENT_IRQ])
                g_mac.callback[MSS_MAC_TXPKTSENT_IRQ]();
    }
    /* Packet received interrupt */
    if((MAC->DMA_IRQ & MAC->DMA_IRQ_MASK) & (1u<<MSS_MAC_RXPKTRCVD_IRQ))
    {
        /* Top-half */
        MAC_rxpkt_handler();
    }
    /* Receiver overflow interrupt */
    if((MAC->DMA_IRQ & MAC->DMA_IRQ_MASK) & (1u<<MSS_MAC_RXOVRFLOW_IRQ))
    {
        /* Call the user's custom handler */
        if( NULL != g_mac.callback[MSS_MAC_TXPKTSENT_IRQ])
                g_mac.callback[MSS_MAC_TXPKTSENT_IRQ]();
    }
    /* Receiver bus error interrupt */
    if((MAC->DMA_IRQ & MAC->DMA_IRQ_MASK) & (1u<<MSS_MAC_RXBUSERR_IRQ))
    {
        /* Call the user's custom handler */
        if( NULL != g_mac.callback[MSS_MAC_TXPKTSENT_IRQ])
                g_mac.callback[MSS_MAC_TXPKTSENT_IRQ]();
    }
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void 
MSS_MAC_enable_irq
(
    uint8_t irq
)
{
    set_bit_reg32(&MAC->DMA_IRQ_MASK, irq);
}
/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void 
MSS_MAC_disable_irq
(
    uint8_t irq
)
{
    clear_bit_reg32(&MAC->DMA_IRQ_MASK, irq);
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void
MSS_MAC_start_tx
(
    void
)
{
    set_bit_reg32( &MAC->DMA_TX_CTRL, DMA_TXENABLE );
    set_bit_reg32( &MAC->CFG1, CFG1_TX_EN);
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void
MSS_MAC_stop_tx
(
    void
)
{
    clear_bit_reg32( &MAC->DMA_TX_CTRL, DMA_TXENABLE );
    clear_bit_reg32( &MAC->CFG1, CFG1_TX_EN);
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void
MSS_MAC_start_rx
(
    void
)
{
    set_bit_reg32( &MAC->DMA_RX_CTRL, DMA_RXENABLE );
    set_bit_reg32( &MAC->CFG1, CFG1_RX_EN);
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void
MSS_MAC_stop_rx
(
    void
)
{
    clear_bit_reg32( &MAC->DMA_RX_CTRL, DMA_RXENABLE );
    clear_bit_reg32( &MAC->CFG1, CFG1_RX_EN);
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
return_t 
MSS_MAC_assign_irq_handler
(
    MAC_irq_t irqnum,
    MAC_callback_t callback    
)
{
    return_t status = FAILED;
    ASSERT( MSS_MAC_INVALID_IRQ > irqnum );
    ASSERT( NULL != callback );

    if( ( MSS_MAC_INVALID_IRQ > irqnum ) && ( NULL != callback ) )
    {
        g_mac.callback[ irqnum ] =  callback ;
        status = SUCCESS;
    }
    return status;
}

/**************************************************************************/
/* Private Function definitions                                           */
/**************************************************************************/  

/**************************************************************************//**
 * MAC_rxpkt_handler()
 * This is default "Receive packet interrupt handler. This function finds the
 * descriptor that received the packet and caused the interrupt.
 * This informs the received packet size to the application and
 * relinquishes the packet buffer from the associated DMA descriptor.
 */
static void 
MAC_rxpkt_handler
(
    void
)
{
    uint32_t rxcnt = 0u;

    /* Read current descriptor from DMA */
    MAC_desc_t * cdesc = (MAC_desc_t *) MAC->DMA_RX_DESC;

    /* Execution comes here because at-least one packet is received.
       Hence rxcnt can be read after freeing one descriptor */
    do
    {
        /* If it is the first descriptor being pointed now, then
           the one which previously received the packet and caused 
           interrupt is the last one in the ring. */
        if( 0u == cdesc->index )
            cdesc = &g_mac.rx_desc_tab[MSS_MAC_RX_RING_SIZE - 1u];
        else
        /* Else the previous one */
            --cdesc;
        /* Packet successfully received indication to user */
        if(cdesc->pkt->lock == LOCK)
        {
            cdesc->pkt->lock = UNLOCK;
            /* Copy no.of bytes received to user pkt excluding FCS */
            cdesc->pkt->size = ( cdesc->pkt_size & DMA_DESC_PKT_SIZE_MASK ) - 4;
            /* Free the packet from driver's access */
            cdesc->pkt = NULL;
        }
        /* Clear the rx packet received interrupt once. If this 
           bit still persists, then another rx packet received 
           interrupt will be generated. Rx count will be decremented.*/
        set_bit_reg32( & MAC->DMA_RX_STATUS, DMA_RXPKTRCVD);
        /* Get the latest current received count */
        rxcnt = ( MAC->DMA_RX_STATUS & DMA_PKTCOUNT_MASK ) >> DMA_PKTCOUNT;
    }while( rxcnt );

    /* Bottom-half : Call the user's custom handler */
    if( NULL != g_mac.callback[MSS_MAC_RXPKTRCVD_IRQ])
            g_mac.callback[MSS_MAC_RXPKTRCVD_IRQ]();
}

/**************************************************************************//**
 * MAC_txpkt_handler()
 * This is default "Transmit packet interrupt handler. This function finds the
 * descriptor that transmitted the packet and caused the interrupt.
 * This relinquishes the packet buffer from the associated DMA descriptor.
 */
static void 
MAC_txpkt_handler
(
    void
)
{
    uint32_t txcnt = 0u;
    /* Read current descriptor from DMA */
    MAC_desc_t * cdesc = (MAC_desc_t *) MAC->DMA_TX_DESC;

    /* Execution comes here because at-least one packet is transmitted.
       Hence txcnt can be read after freeing one descriptor */
    do
    {
        /* If it is the first descriptor being pointed now, then
           the one which previously transmitted the packet and caused 
           interrupt is the last one in the ring. */
        if( 0u == cdesc->index )
            cdesc = &g_mac.tx_desc_tab[MSS_MAC_TX_RING_SIZE - 1u];
        else
        /* Else the previous one */
            --cdesc;
        /* Packet successfully transmitted indication to user */
        if( LOCK == cdesc->pkt->lock)
        {
            cdesc->pkt->lock = UNLOCK;
            /* Free the packet from driver's access */
            cdesc->pkt = NULL;
        }
        /* Clear the tx packet sent interrupt once. If this 
           bit still persists, then another tx packet is awaiting 
           for transmission and interrupt will be generated. */
        set_bit_reg32( & MAC->DMA_TX_STATUS, DMA_TXPKTSENT);
        /* Get the latest current received count */
        txcnt = ( cdesc->pkt_size & DMA_PKTCOUNT_MASK );
    }while( txcnt );

    /* Bottom-half : Call the user's custom handler */ 
    if( NULL != g_mac.callback[MSS_MAC_TXPKTSENT_IRQ])
            g_mac.callback[MSS_MAC_TXPKTSENT_IRQ]();
}

/**************************************************************************//**
 * 
 */                                  
static void 
MAC_desc_ring_init
(
    MAC_desc_t * ring, 
    MAC_io_t io
)
{
    uint32_t l = 0u;
    uint32_t size = 0u;

    (TX == io) ? ( size = MSS_MAC_TX_RING_SIZE) : (size = MSS_MAC_RX_RING_SIZE);
    for(l=0; l<size; l++)    
    {
        /* Set packet start address to NULL */
        ring[l].pkt_start_addr = (uint32_t)NULL;
        /* 
         * Set empty flag indicating invalid data in case of Tx descriptor.
         * clear empty flag indicating full buffer in case of Rx descriptor.
         * This disables the DMA from transferring packets to and from FIFO 
         * in case of Tx descriptor and Rx descriptors respectively.
         */
        (TX == io) ? SET_DESC_EMPTY_FLAG(ring[l].pkt_size) : 
                            CLEAR_DESC_EMPTY_FLAG(ring[l].pkt_size);
        /* Every descriptor links to next descriptor */
        ring[l].next_desriptor = (uint32_t)(&ring[l+1]);
        ring[l].pkt = NULL;
        ring[l].index = l;
    }
    /* Complete the ring by connecting tail and head */
    ring[l-1u].next_desriptor = (uint32_t)(&ring[0]);
    
}

/**************************************************************************//**
 * 
 */
static return_t
MAC_assign_station_addr
(
    const uint8_t *mac_addr
)
{
    return_t status = FAILED;
    uint32_t temp;

    ASSERT( NULL!= mac_addr );            

    if( NULL != mac_addr )
    {
    	temp = ((uint32_t)mac_addr[0]) << 24;
	    temp |= ((uint32_t)mac_addr[1])<< 16;
	    temp |= ((uint32_t)mac_addr[2])<< 8;
	    temp |= ((uint32_t)mac_addr[3]);
	    MAC->STATION_ADDRESS1 = temp;
	    temp = ((uint32_t)mac_addr[4]) << 24;
	    temp |= ((uint32_t)mac_addr[5])<< 16;
	    MAC->STATION_ADDRESS2 = temp;
        status = SUCCESS;
    }
    return status;
}

/***************************************************************************//**
 * MAC_auto_link_setup()
 * This function is called when MSS_MAC_init() is function is called with a
 * configuration having auto-negotiation enabled. This function calls the 
 * PHY interface functions to auto-negotiate link parameters / PHY capabilites.
 * Returns 1 on successful execution. Returns 0 otherwise.
 */
static return_t
MAC_auto_link_setup
(
    MAC_cfg_t * cfg
)
{
    return_t status = FAILED;

    /* Configure for auto-negotiation.
       If auto-negotiation is enabled, then wait till it
       is completed. */
    status = PHY_device[cfg->phy].config_aneg( cfg );

    /* Read the PHY status after auto-negotiation is completed.
       And change the elements in cfg as per the link 
       capabilities detected */
    if( FAILED != status )
        status = PHY_device[cfg->phy].read_status( cfg );

    return status;    
}

/***************************************************************************//**
 * MAC_memcpy()
 * Thsi copies n bytes from memory area src to memory area dest.
 * The memory areas should not overlap.
 *
 * @return          a pointer to the memory area dest.
 */
static uint8_t * MAC_memcpy(uint8_t *dest, const uint8_t *src, uint32_t n)
{
    uint8_t *d = dest;

    while( n > 0 ) {
    	n--;
        d[n] = src[n];
    }

    return dest;
}

#ifdef __cplusplus
}
#endif

/******************************** END OF FILE ******************************/
