/*******************************************************************************
 * (c) Copyright 2012 Microsemi Corporation.  All rights reserved.
 *
 *  Microsemi Smartfusion2 USB device logical driver layer Implementation.	
 *  
 *
 *  This file implements MSD class specific initialization and request handling.
 *
 * SVN $Revision: 5027 $
 * SVN $Date: 2013-02-01 19:06:39 +0530 (Fri, 01 Feb 2013) $
 */
#include "mss_usb_device_msd.h"
#include "mss_usb_device.h"
#include "../../CMSIS/mss_assert.h"

#ifdef __cplusplus
extern "C" {
#endif 

#define MSD_TX_EP_FIFO_SIZE                         512
#define MSD_RX_EP_FIFO_SIZE                         512

  /*Values should be same as described in ep descriptors*/
#define MSD_TX_EP_MAX_PKT_SIZE                      512  
#define MSD_RX_EP_MAX_PKT_SIZE                      512  

#define G_BOT_CBW_INIT();		                    g_bot_cbw.signature = 0;\
													g_bot_cbw.tag = 0;\
													g_bot_cbw.xfr_length = 0;\
													g_bot_cbw.flags = 0;\
													g_bot_cbw.lun = 0;\
													g_bot_cbw.cmd_length = 0;\
													g_bot_cbw.cmd_block[0]=(uint8_t)0;\
													g_bot_cbw.cmd_block[1]=(uint8_t)0;\
													g_bot_cbw.cmd_block[2]=(uint8_t)0;\
													g_bot_cbw.cmd_block[3]=(uint8_t)0;\
													g_bot_cbw.cmd_block[4]=(uint8_t)0;\
													g_bot_cbw.cmd_block[5]=(uint8_t)0;\
													g_bot_cbw.cmd_block[6]=(uint8_t)0;\
													g_bot_cbw.cmd_block[7]=(uint8_t)0;\
													g_bot_cbw.cmd_block[8]=(uint8_t)0;\
													g_bot_cbw.cmd_block[9]=(uint8_t)0;\
													g_bot_cbw.cmd_block[10]=(uint8_t)0;\
													g_bot_cbw.cmd_block[11]=(uint8_t)0;\
													g_bot_cbw.cmd_block[12]=(uint8_t)0;\
													g_bot_cbw.cmd_block[13]=(uint8_t)0;\
													g_bot_cbw.cmd_block[14]=(uint8_t)0;\
													g_bot_cbw.cmd_block[15]=(uint8_t)0;

/***************************************************************************//**
  Local functions used by MSD class driver.
 */
mss_usb_msd_csw_status_t MSS_USB_MSD_process_cbw( mss_usb_msd_cbw_t * cbw );
void MSS_USB_MSD_prepare_sense_data( uint8_t lun, uint8_t skey, uint8_t asc);
void MSS_USB_MSD_abort_cbw( uint8_t lun );
void MSS_USB_MSD_BOT_state_machine( uint8_t status, uint32_t rx_count );
void MSS_USB_MSD_send_data( uint8_t* buf, uint32_t len );
void MSS_USB_MSD_receive_data( uint8_t* buf, uint32_t len );
void MSS_USB_MSD_send_csw( mss_usb_msd_csw_status_t csw_status);
mss_usb_msd_csw_status_t MSS_USB_MSD_process_read_10( void );
mss_usb_msd_csw_status_t MSS_USB_MSD_process_write_10( void );

/***************************************************************************//**
  Local implementations of Call-back functions used by USB Device mode driver.
 */
uint8_t* MSS_USB_MSD_get_descriptor_cb(uint8_t recepient, uint8_t type, uint32_t* length);
uint8_t MSS_USB_MSD_init_cb(uint8_t cfgidx);
uint8_t MSS_USB_MSD_deinit_cb(uint8_t cfgidx);
uint8_t MSS_USB_MSD_data_in_cb(mss_usb_ep_num_t num, uint8_t status);
uint8_t MSS_USB_MSD_data_out_cb(mss_usb_ep_num_t num, uint8_t status, uint32_t rx_count);
uint8_t MSS_USB_MSD_process_request_cb(uint8_t** buf_pp,
                                      uint8_t request,
                                      uint32_t* length,
                                      uint8_t* data_buf_p,
                                      uint32_t data_len);

/*******************************************************************************
 Global variables used by MSD class driver.
 */

/*Buffer for responding to GET_MAX_LUN*/
uint8_t get_max_lun_response[] = {0x00u};

/*
  Buffer for responding to READ_FORMAT CAPACITIES command. Values will be provided 
  by Apps.
  Driver Does not store Values for All the LUNs.
  Values of the The LUN on which last USB_MSD_SCSI_READ_FORMAT_CAPACITIES command
  was executed will be there in this buffer.
*/
uint8_t format_capacity_list[12] = {0};

/*
  Buffer for responding to READ_CAPACITY_10 command. Values will be provided by Apps.
  Driver retains values for all the supported LUN and uses them to calculate valid
  address range for that LUN.
*/
mss_usb_msd_lun_capacity_t lun_capacity[4] = {{0},{0},{0},{0}};

/*
  Buffer to respond to USB_MSD_SCSI_MODE_SENSE_6 command.
  Values are generated by driver.Same values are returned for all supported LUNs.
*/
uint8_t mode_sense_response[12] =
{
	0x0Cu, 0x00u, 0x00u, 0x00u, 0x00u, 0x00u,
	0x00u, 0x00u, 0x00u, 0x00u, 0x00u, 0x00u
};

/* Definition for Flash drive application call-backs. */
mss_usb_msd_media_t *g_mss_usb_msd_media_ops;

/*
  Global variables used for multi-transfer OUT transactions
*/
uint8_t* g_rx_buf_ptr;
uint32_t g_rx_buf_len = 0u;
uint32_t g_rx_current_lba_addr = 0u;

/*
  Global variables used for multi-transfer IN transactions
*/
uint8_t* g_tx_buf_ptr = 0u;
uint32_t g_tx_buf_len = 0u;
uint32_t g_tx_current_lba_addr = 0u;

/*
  The g_bot_state is used to store the current state of the driver during
   Bulk-only Transport (BOT) transaction.
  The drivers respose to the host request depends on this state.
*/
mss_usb_msd_bot_state_t g_bot_state;
/*
  This veriable track the current state of the MSD Class driver.
*/
usb_msd_state_t g_msd_driver_state = USB_MSD_NOT_CONFIGURED;
/*
  The g_sense_response is used to respond to REQUEST_SENSE request.
  Value of this variable is changed depending on the type of failure happened while
   responding to the host request.
*/
scsi_request_sense_response_t g_sense_response;

/*
  The g_current_command_csw structure is used to provide the status of the command.
  The Values of the elements of this structure are updated according to the state
  of the driver,success/Failure of the command and type of failure.
*/
mss_usb_msd_csw_t g_current_command_csw = {USB_MSD_BOT_CSW_SIGNATURE,
                                           0,
                                           0,
                                           SCSI_COMMAND_PASSED };

 /*
  The g_bot_cbw is used to store the current CBW sent by the the host as a 
  BOT request.
 */
mss_usb_msd_cbw_t g_bot_cbw;

 /* The g_bot_events is used to know the current event in the BOT transfer */
mss_usb_msd_bot_events_t g_bot_events = BOT_EVENT_IDLE;

mss_usb_class_cb_t usb_msd_class_cb = {MSS_USB_MSD_init_cb,
                                       MSS_USB_MSD_deinit_cb,
                                       MSS_USB_MSD_get_descriptor_cb,
                                       MSS_USB_MSD_process_request_cb,
                                       MSS_USB_MSD_data_in_cb,
                                       MSS_USB_MSD_data_out_cb,
                                       0,
                                       0 };
/*
Configuration descriptor and sub-ordinate descriptors to enumerate the USB device
as Mass Storage class Device by host.
*/
uint8_t msd_configuration_descriptor[FULL_CONFIG_DESCR_LENGTH] =
{
    /*----------------------- Configuration Descriptor -----------------------*/
    CONFIGURATION_DESCR_LENGTH,                     /* bLength */
    USB_CONFIGURATION_DESCRIPTOR_TYPE,              /* bDescriptorType */
    FULL_CONFIG_DESCR_LENGTH,                       /* wTotalLength LSB */
    0x00u,                                          /* wTotalLength MSB */
    0x01u,                                          /* bNumInterfaces */
    0x01u,                                          /* bConfigurationValue */
    0x04u,                                          /* iConfiguration */
    0xC0u,                                          /* bmAttributes */
    0x32u,                                          /* bMaxPower */				
    /*------------------------- Interface Descriptor -------------------------*/
    INTERFACE_DESCR_LENGTH,                         /* bLength */
    USB_INTERFACE_DESCRIPTOR_TYPE,                  /* bDescriptorType */
    0x00u,                                          /* bInterfaceNumber */
    0x00u,                                          /* bAlternateSetting */
    0x02u,                                          /* bNumEndpoints */
    0x08u,                                          /* bInterfaceClass */
    0x06u,                                          /* bInterfaceSubClass */
    0x50u,                                          /* bInterfaceProtocol */
    0x05u,                                          /* bInterface */
	/*------------------------- IN Endpoint Descriptor --------------------------*/
	ENDPOINT_DESCR_LENGTH,				            /* bLength */
	USB_ENDPOINT_DESCRIPTOR_TYPE,		            /* bDescriptorType */
	0x81u,								            /* bEndpointAddress */		
	0x02u,								            /* bmAttributes */
	0x00u,								            /* wMaxPacketSize LSB */
	0x02u,								            /* wMaxPacketSize MSB */
	0xFFu,								            /* bInterval */ 	        
    /*------------------------- OUT Endpoint Descriptor --------------------------*/
    ENDPOINT_DESCR_LENGTH,                          /* bLength */
    USB_ENDPOINT_DESCRIPTOR_TYPE,                   /* bDescriptorType */
    0x01u,                                          /* bEndpointAddress */		
    0x02u,                                          /* bmAttributes */
    0x00u,                                          /* wMaxPacketSize LSB */
    0x02u,                                          /* wMaxPacketSize MSB */
    0xFFu                                           /* bInterval */		        /*Max NAK rate*/
};

void
MSS_USB_MSD_init
(
    mss_usb_msd_media_t* media_ops
)
{
    g_mss_usb_msd_media_ops = media_ops;
    MSS_USB_device_set_class_cb_handler(&usb_msd_class_cb);
}

usb_msd_state_t
MSS_USB_MSD_get_state
(
    void
)
{
    return g_msd_driver_state;
}
/***************************************************************************//**
 * returns the configuration descriptor requested by Host.
 */
uint8_t*
MSS_USB_MSD_get_descriptor_cb
(
	uint8_t recepient,
	uint8_t type,
	uint32_t* length
)
{
	if( ( recepient == USB_STD_REQ_RECIPIENT_DEVICE ) && ( type ==USB_CONFIGURATION_DESCRIPTOR_TYPE ) )
    {
	   *length = sizeof( msd_configuration_descriptor );
        return( msd_configuration_descriptor );
	}
	else if( recepient == USB_STD_REQ_RECIPIENT_ENDPOINT )    /*Need index(EP Num) here*/
	{
        /*Do nothing*/
	}
	else if( recepient == USB_STD_REQ_RECIPIENT_INTERFACE )     /*Need index(intrfs num) here*/
	{
        /*Do nothing*/
	}
    else
    {
        /*Do nothing*/
    }
    return FAIL;
}

/***************************************************************************//**
 MSS_USB_MSD_init_cb() call-back is called by USB Device mode driver on receiving 
 SET_CONFIGURATION command. The MSD class specific configurations are performed
 by this function.
*/
uint8_t
MSS_USB_MSD_init_cb
(
	uint8_t cfgidx
)
{
    g_bot_state = BOT_IDLE;
	MSS_USB_device_rx_ep_configure(MSS_USB_RX_EP_1,
                                   0x100,
                                   MSD_RX_EP_FIFO_SIZE,
                                   MSD_RX_EP_MAX_PKT_SIZE,
                                   1,
                                   DMA_ENABLE,
                                   MSS_USB_DMA_CHANNEL1,
                                   MSS_USB_XFR_BULK,
                                   NO_ZLP_TO_XFR);

	MSS_USB_device_rx_ep_read_prepare(MSS_USB_RX_EP_1,
                                      (uint8_t*)&g_bot_cbw,
                                      31 );

    MSS_USB_device_tx_ep_configure(MSS_USB_TX_EP_1,
                                   0x200,
                                   MSD_TX_EP_FIFO_SIZE,
                                   MSD_TX_EP_MAX_PKT_SIZE,
                                   1,
                                   DMA_ENABLE,
                                   MSS_USB_DMA_CHANNEL2,
                                   MSS_USB_XFR_BULK,
                                   NO_ZLP_TO_XFR);
	g_msd_driver_state = USB_MSD_CONFIGURED;
    return 1;
}

/***************************************************************************//**
 MSS_USB_MSD_deinit_cb() call-back is called by USB Device mode driver on receiving 
 a command to clear the configuration.
 */
uint8_t
MSS_USB_MSD_deinit_cb
(
	uint8_t cfgidx
)
{
  	g_msd_driver_state = USB_MSD_NOT_CONFIGURED;
	return 1;
}

/***************************************************************************//**
 MSS_USB_MSD_process_request_cb() call-back function Process the MSD class requests. 
 */
uint8_t
MSS_USB_MSD_process_request_cb
(
	uint8_t** buf_pp,
	uint8_t request,
	uint32_t* length,
	uint8_t* data_buf_p,
	uint32_t data_len
)
{
    switch( request )
    {
        case USB_MSD_BOT_REQ_GET_MAX_LUN:

            /*Return the max INDEX, not the actual number of LUNs*/
            ASSERT( g_mss_usb_msd_media_ops->msd_media_get_max_lun !=0 );
            get_max_lun_response[0] = ( g_mss_usb_msd_media_ops->msd_media_get_max_lun() - 1 );
            *length = sizeof( get_max_lun_response );
            *buf_pp = (uint8_t*)&get_max_lun_response;

		case USB_MSD_BOT_REQ_BMS_RESET:
			/*
			LVp:62 - prepare for next CBW. Dont change the stall
            and Data toggle bit on Bulk EPs.
			*/
			g_bot_events = BOT_EVENT_IDLE;
			g_bot_state = BOT_IDLE;
			g_current_command_csw.data_residue = 0u;
			g_current_command_csw.tag = 0u;
			g_current_command_csw.status = SCSI_COMMAND_FAILED;
			G_BOT_CBW_INIT();
			break;

        default:
            length = 0u;
            break;
    }
    return SUCCESS;
}

/***************************************************************************//**
 MSS_USB_MSD_data_in_cb() call-back function is called by USB Device mode driver 
 on completion of the Current Data Transmissions (IN Transaction) which was
 previously initiated using MSS_USB_device_tx_ep_configure().
 */
uint8_t MSS_USB_MSD_data_in_cb
(
    mss_usb_ep_num_t num,
    uint8_t status
)
{
	if( (status & 0x1Fu) > 0x01u)
	{
		g_bot_events = BOT_EVENT_DATAIN_ERROR;
	}
	else
	{
		g_bot_events = BOT_EVENT_DATAIN;
	}
	MSS_USB_MSD_BOT_state_machine( status, 0 );

    return 1;
}

/***************************************************************************//**
 MSS_USB_MSD_data_out_cb() call-back function is called by USB Device mode driver 
 on completion of data receiption. USB Device mode driver must have been previously
 prepared for this data reception using MSS_USB_device_rx_ep_read_prepare()
 */
uint8_t
MSS_USB_MSD_data_out_cb
(
	mss_usb_ep_num_t num,
	uint8_t status,
	uint32_t rx_count
)
{
    /*
    4.3.5.6
    xfr length > Allocation length field ==> Terminate with CHECK CONDITION status,
    sense key = ILLEGAL_REQUEST, additional sense code = INVALID FIELD in CDB
    when NACA bit = 0 ==> ACA condition is not set = command abort should be done
    as per sam-5, 5.8.2
    */
    if( (status & 0x1Fu) > 0x01u)
    {
        g_bot_events = BOT_EVENT_DATAOUT_ERROR;
    }
    else
    {
        g_bot_events = BOT_EVENT_DATAOUT;
    }

    MSS_USB_MSD_BOT_state_machine( status, rx_count );

    return 1;
}

/***************************************************************************//**
 MSS_USB_MSD_BOT_state_machine() function is the state machine for BOT transfers.
 */
void
MSS_USB_MSD_BOT_state_machine
(
	uint8_t status,
	uint32_t rx_count
)
{
    mss_usb_msd_csw_status_t command_status;

    switch (g_bot_state)
	{
		case BOT_IDLE:
			if( g_bot_events == BOT_EVENT_DATAOUT )
			{
                command_status = MSS_USB_MSD_process_cbw( &g_bot_cbw );

                g_current_command_csw.status = command_status;

				if( g_current_command_csw.status == SCSI_COMMAND_FAILED )
                {
                   if( g_bot_cbw.xfr_length == 0u )
                    {
                        MSS_USB_MSD_send_csw( command_status );
                        MSS_USB_MSD_prepare_sense_data(0,
                                                       SC_ILLEGAL_REQUEST,
                                                       ASC_INVALID_CDB );
                    }
                    else if( g_bot_cbw.flags == 0x80u ) /*In Transaction*/
                    {
                        /*Send ZLP if data was requested*/
                        /*Dummy buffer expected by Windows host*/
                        MSS_USB_MSD_send_data(((uint8_t*)" "), 0u);
                    }
                    else
                    {
                        MSS_USB_MSD_send_csw( command_status );
                        MSS_USB_MSD_prepare_sense_data(0,
                                                       SC_ILLEGAL_REQUEST,
                                                       ASC_INVALID_CDB );
                        g_current_command_csw.data_residue = 0u;
                    }
                }
                else
                {
                    /*Command with no Data transfer phase*/
                    if( g_bot_cbw.xfr_length ==  0u )
                    {
                        /*There is no data phase for the command*/
                        MSS_USB_MSD_send_csw( command_status );
                    }
                    /*Command with either IN or OUT Data transfer phase*/
                    else if( g_bot_cbw.xfr_length >  0u )
                    {
                        /*
                         Flags.bit7 indicates the directiont of data transport
                         phase.
                         0x01 indicates IN transfers is expected.
                        */
                        if( g_bot_cbw.flags == 0x80 ) /*In Transaction*/
                        {
                            /*READ_FORMAT_CAPABILITIES is expecting 252 bytes of data.*/
                            /*Inquiry command attemts to receive different length everytime.*/
                            if( g_tx_buf_len <= g_bot_cbw.xfr_length )
                            {
                                MSS_USB_MSD_send_data( g_tx_buf_ptr, g_tx_buf_len );
                            }
                            else if ( g_tx_buf_len > g_bot_cbw.xfr_length )
                            {
                                MSS_USB_MSD_send_csw( SCSI_COMMAND_PHASE_ERR );
                                MSS_USB_MSD_prepare_sense_data(0,
                                                               SC_ILLEGAL_REQUEST,
                                                               ASC_INVALID_CDB );
                            }
                            else
                            {
                                MSS_USB_MSD_send_csw( SCSI_COMMAND_FAILED );
                                MSS_USB_MSD_prepare_sense_data( 0,
                                                                SC_ILLEGAL_REQUEST,
                                                                ASC_INVALID_CDB);
                            }
                        }
                        else if( g_bot_cbw.flags == 0x00 ) /*Out transaction, prepare to receive data*/
                        {
                            if( g_rx_buf_len <= g_bot_cbw.xfr_length )
                            {
                                MSS_USB_MSD_receive_data( g_rx_buf_ptr, g_rx_buf_len );
                            }
                            else if( g_rx_buf_len > g_bot_cbw.xfr_length )
                            {
                                MSS_USB_MSD_send_csw( SCSI_COMMAND_PHASE_ERR );
                                MSS_USB_MSD_prepare_sense_data( 0, SC_ILLEGAL_REQUEST,ASC_INVALID_CDB );
                            }
                            else
                            {
                                MSS_USB_MSD_send_csw( SCSI_COMMAND_FAILED );
                                MSS_USB_MSD_prepare_sense_data(0,
                                                               SC_ILLEGAL_REQUEST,
                                                               ASC_INVALID_CDB );
                            }
                        }
                    }
                }
			}
			else
			{
				/*No other event is expected here.*/
				MSS_USB_MSD_abort_cbw( 0 );
			}
			break;
		case BOT_DATA_IN:
          	if( g_bot_events == BOT_EVENT_DATAIN ) /* Data TX from Device*/
			{
                if( (g_current_command_csw.status == SCSI_COMMAND_FAILED ) ||
                    (g_current_command_csw.status == SCSI_COMMAND_PHASE_ERR ) )
                {
                    g_rx_buf_ptr = 0;
                    g_rx_buf_len = 0;
                    g_rx_current_lba_addr = 0;
                    MSS_USB_MSD_send_csw( g_current_command_csw.status );
                }
                else if(g_current_command_csw.status == SCSI_COMMAND_PASSED)
                {
                    if( g_current_command_csw.data_residue == 0 )
                    {
                        g_tx_buf_ptr = 0;
                        g_tx_buf_len = 0;
                        g_rx_current_lba_addr = 0;
                        MSS_USB_MSD_send_csw( g_current_command_csw.status );
                    }
                    else
                    {
                        /* 
                         Control reaches here because previous App read operation 
                         returned less
                         Data than requested by host.Read rest of the data now.
                        */
                        uint32_t actual_read_len;
                        uint8_t* buf;

                        g_tx_current_lba_addr += g_tx_buf_len;
                        actual_read_len = g_mss_usb_msd_media_ops->msd_media_read(g_bot_cbw.lun,
                                                                                &buf,
                                                                                g_tx_current_lba_addr,
                                                                                g_current_command_csw.data_residue );
                        /*App should not read more data than asked for*/
                        if ( actual_read_len > g_current_command_csw.data_residue )
                        {
                            ASSERT(0);
                        }
                        else
                        {
                            g_tx_buf_ptr = buf;
                            g_tx_buf_len = actual_read_len;
                        }

                        MSS_USB_MSD_send_data( buf, actual_read_len );
                    }
                }
			}
			else if( g_bot_events == BOT_EVENT_DATAIN_ERROR )
			{
                /* Underrun error */
                if( status & 0x01 )
                {
                    status &= ~0x01;
                    G_BOT_CBW_INIT();
                	g_current_command_csw.data_residue = 0u;
                	g_tx_buf_ptr = (uint8_t)0;
                	g_tx_buf_len = 0;
					MSS_USB_device_rx_ep_read_prepare(MSS_USB_RX_EP_1,
                                                      (uint8_t*)&g_bot_cbw,
                                                      31 );
                }
                /* Stall error */
                if( status&0x02 )
                {
                    status &= ~0x02;
                }
				MSS_USB_MSD_abort_cbw( 0 );
   				MSS_USB_MSD_prepare_sense_data(0,
                                               SC_ILLEGAL_REQUEST,
                                               ASC_INVALID_CDB );
            }
		break;
		case BOT_DATA_OUT:
			if( g_bot_events == BOT_EVENT_DATAOUT )
			{
                if(g_current_command_csw.status == SCSI_COMMAND_FAILED)
                {
                    MSS_USB_core_tx_ep_stall( MSS_USB_TX_EP_1 );
                }
                else
                {
                    /*Write the received data on flash*/
                    if( rx_count == g_rx_buf_len )
                    {
                        /*
                        TODO make the return value as SUCCESS/Fail,
                        send status accordingly
                        */
                        g_mss_usb_msd_media_ops->msd_media_write_ready(g_bot_cbw.lun,
                                                                     g_rx_current_lba_addr,
                                                                     g_rx_buf_len );

                        g_current_command_csw.data_residue -= g_rx_buf_len;
                    }

                    /*
                    data_residue == 0 means we are done with current Data phase of
                    OUT transaction, send the status.
                    */
                    if( g_current_command_csw.data_residue == 0u )
                    {
                        g_rx_buf_ptr = 0; // Data end.
                        g_rx_buf_len = 0;
                        g_rx_current_lba_addr = 0;
                        command_status = SCSI_COMMAND_PASSED;
                        MSS_USB_MSD_send_csw( SCSI_COMMAND_PASSED );
                    }
                    else
                    {
                        uint32_t app_write_len;
                        uint8_t* write_buf;

                        /*
                        Data phase of current OUT transaction is still in progress.
                        Prepare to write more data on flash
                        Additionally, make sure that the driver has read the same
                        amount of data that was requested.
                        */
                        if( rx_count == g_rx_buf_len )
                        {
                            g_rx_current_lba_addr += g_rx_buf_len;
                        }
                        else
                        {
                            ASSERT( 0 );
                        }

                        write_buf = g_mss_usb_msd_media_ops->msd_media_acquire_write_buf(g_bot_cbw.lun,
                                                                                       g_rx_current_lba_addr,
                                                                                       &app_write_len );

                        /*
                        If application buffer is not big enough to handle all requested
                        data from Host, then read amount of data that the application
                        can handle at the moment
                        */
                        if( app_write_len < g_current_command_csw.data_residue )
                        {
                            g_rx_buf_len = app_write_len;
                        }
                        else
                        {
                            g_rx_buf_len = g_current_command_csw.data_residue;
                        }

                        g_rx_buf_ptr = write_buf;

                        MSS_USB_MSD_receive_data (g_rx_buf_ptr, g_rx_buf_len );
                    }
                }
    		}
			else if( g_bot_events == BOT_EVENT_DATAOUT_ERROR )
			{
                /* Underrun Error */
                if( status&0x01 )
                {
                    status &= ~0x01;
                    G_BOT_CBW_INIT();
                	g_current_command_csw.data_residue = 0;
                	g_tx_buf_ptr = (uint8_t*)0;
                	g_tx_buf_len = 0u;
					MSS_USB_device_rx_ep_read_prepare(MSS_USB_RX_EP_1,
                                                      (uint8_t*)&g_bot_cbw,
                                                      31 );
                }

                /* Stall Error*/
                if( status&0x02 )
                {
                    status &= ~0x02;
                }
				MSS_USB_MSD_abort_cbw( 0 );
   				MSS_USB_MSD_prepare_sense_data(0,
                                               SC_ILLEGAL_REQUEST,
                                               ASC_INVALID_CDB );
            }
		break;
		case BOT_SEND_STATUS:
			if( g_bot_events == BOT_EVENT_DATAIN )
			{
    			g_bot_state = BOT_IDLE;

                /*Stall the IN EP if we transmitted less data than expecetd by host*/
                if( ( g_current_command_csw.data_residue ) &&
                    ( g_bot_cbw.xfr_length ) &&
                    ( g_bot_cbw.flags == 0x80 ) )
                {
                    MSS_USB_core_tx_ep_stall( MSS_USB_TX_EP_1 );
                }
                g_tx_buf_ptr = 0;       /* Data end.*/
				g_tx_buf_len = 0u;
				g_rx_buf_ptr = 0;
				g_rx_buf_len = 0u;
			}
			else
			{
				/*No other event is expected here.*/
				MSS_USB_MSD_abort_cbw( SCSI_COMMAND_FAILED );
				MSS_USB_MSD_prepare_sense_data(0,
                                               SC_ILLEGAL_REQUEST,
                                               ASC_INVALID_CDB );
			}
		break;

        default:
            ASSERT( 0 );
		break;
	}

	g_bot_events = BOT_EVENT_IDLE;
}

/***************************************************************************//**
 MSS_USB_MSD_process_cbw() function decodes current command sent by host in CBW,
 and Processes the command. If the command was valid and meaningful, it asks for
 required information from application and returns the status. If the Command
 was invalid or is not processed by this driver, it returns the Failure status.
 */
mss_usb_msd_csw_status_t
MSS_USB_MSD_process_cbw
(
    mss_usb_msd_cbw_t * cbw
)
{
    uint8_t cbw_command = cbw->cmd_block[0];
	mss_usb_msd_csw_status_t command_status = SCSI_COMMAND_FAILED;


	/*
	 Check for valid cbw
	 1) CBW received after csw or reset
	 2) cbw is 32 bytes
	 3) Signature is correct*/

	/*
	 CBW is meaninggful if all reserved bytes are zero
	 Lun value is supported
	 CBWCBLength and cbwcb are valid for the interface subclass.
     No. Max LUN index provided by App in get_max_lun reques.
	*/
	if(	( cbw->signature != USB_MSD_BOT_CBW_SIGNATURE ) ||
		( cbw->lun > get_max_lun_response[0] ) ||
		( cbw->cmd_length < 1 ) ||
		( cbw->cmd_length > 16 ) )
	{
		MSS_USB_MSD_prepare_sense_data(cbw->lun,
                                       SC_ILLEGAL_REQUEST,
                                       ASC_INVALID_CDB );

		MSS_USB_MSD_abort_cbw( cbw->lun );
	}
	else
    {
        ASSERT( g_current_command_csw.data_residue == 0 );
        g_current_command_csw.tag = cbw->tag;

	    switch( cbw_command )
        {
            case USB_MSD_SCSI_INQUIRY:
              {
                /*TODO: needs to be updated for multiple luns*/
				if( g_mss_usb_msd_media_ops->msd_media_inquiry != 0 )
				{
					g_tx_buf_ptr = g_mss_usb_msd_media_ops->msd_media_inquiry(cbw->lun,
                                                                            &g_tx_buf_len );

                    g_current_command_csw.data_residue = g_tx_buf_len;
					command_status = SCSI_COMMAND_PASSED;
				}
              }
            break;
            case USB_MSD_SCSI_READ_CAPACITY_10: /*returns last LBA and block size*/
              {
                /*
                 Returns last block addr and block size,
                 Last_blk_lba = no_of_blocks-1
                */
                if( g_mss_usb_msd_media_ops->msd_media_get_capacity != 0 )
                {
                      uint32_t no_of_blocks, block_size;
                      g_mss_usb_msd_media_ops->msd_media_get_capacity(cbw->lun,
                                                                    (uint32_t*)&no_of_blocks,
                                                                    (uint32_t*)&block_size );

                      lun_capacity[cbw->lun].last_lba_msb = (uint8_t)( ( no_of_blocks - 1 ) >> 24 );
                      lun_capacity[cbw->lun].last_lba_2 = (uint8_t)( ( no_of_blocks - 1 ) >> 16 );
                      lun_capacity[cbw->lun].last_lba_1 = (uint8_t)( ( no_of_blocks - 1 ) >> 8 );
                      lun_capacity[cbw->lun].last_lba_lsb = (uint8_t)( ( no_of_blocks - 1 ) );

			          lun_capacity[cbw->lun].blk_sz_len = block_size;
                      lun_capacity[cbw->lun].block_size_msb = (uint8_t)( block_size >> 24 );
                      lun_capacity[cbw->lun].block_size_2 = (uint8_t)( block_size >> 16 );
                      lun_capacity[cbw->lun].block_size_1 = (uint8_t)( block_size >> 8 );
                      lun_capacity[cbw->lun].block_size_lsb = (uint8_t)( block_size );

                      g_tx_buf_ptr = (uint8_t*)&lun_capacity[cbw->lun];
                      g_tx_buf_len = 8u;
                      command_status = SCSI_COMMAND_PASSED;

                      /*Windows expects different length of data everytime*/
                      g_current_command_csw.data_residue = g_tx_buf_len;
                  }
              }
                break;
        	case USB_MSD_SCSI_READ_FORMAT_CAPACITIES:/*Returns no. of blocks and block size*/
              {
                  uint32_t no_of_blocks, block_size;

                  g_mss_usb_msd_media_ops->msd_media_get_capacity(cbw->lun,
                                                                (uint32_t*)&no_of_blocks,
                                                                (uint32_t*)&block_size );

                  format_capacity_list [4] = (uint8_t)( no_of_blocks >> 24 );
                  format_capacity_list [5] = (uint8_t)( no_of_blocks >> 16 );
                  format_capacity_list [6] = (uint8_t)( no_of_blocks >> 8  );
                  format_capacity_list [7] = (uint8_t)( no_of_blocks );

                  /* format_capacity_list [8] = Descriptor Code*/
                  format_capacity_list [9]  = (uint8_t)( block_size >> 16 );
                  format_capacity_list [10] = (uint8_t)( block_size >> 8  );
                  format_capacity_list [11] = (uint8_t)( block_size );

                  g_tx_buf_ptr = (uint8_t*)format_capacity_list;
                  g_tx_buf_len = sizeof( format_capacity_list );
                  g_current_command_csw.data_residue = g_tx_buf_len;
                  command_status = SCSI_COMMAND_PASSED;
              }
              break;
              case USB_MSD_SCSI_READ_10:
                  command_status = MSS_USB_MSD_process_read_10();
              break;

            case USB_MSD_SCSI_MODE_SENSE_6: /*TODO: needs to be updated for multiple luns*/
                  g_tx_buf_ptr = mode_sense_response;
                  g_tx_buf_len = sizeof( mode_sense_response );
                  g_current_command_csw.data_residue = g_tx_buf_len;
                  command_status = SCSI_COMMAND_PASSED;
                break;

            case USB_MSD_SCSI_REQUEST_SENSE: /*TODO: needs to be updated for multiple luns*/
				/*SPC-4, table 39*/
				command_status = SCSI_COMMAND_PASSED;
				g_tx_buf_ptr = (uint8_t*)&g_sense_response;
				g_tx_buf_len = sizeof(g_sense_response);
                g_current_command_csw.data_residue = g_tx_buf_len;
                break;

			case USB_MSD_SCSI_MODE_SELECT_6:    /*TODO: needs to be updated for multiple luns*/
                break;

            case USB_MSD_SCSI_TEST_UNIT_READY:  /*TODO: needs to be updated for multiple luns*/
				command_status = SCSI_COMMAND_PASSED;
				g_tx_buf_ptr = (uint8_t)0;
				g_tx_buf_len = 0u;
                g_current_command_csw.data_residue = g_tx_buf_len;
                break;

            case USB_MSD_SCSI_WRITE_10:
				command_status = MSS_USB_MSD_process_write_10();
                break;

            case USB_MSD_SCSI_PREVENT_ALLW_MDM_RMVL:    /*TODO: needs to be updated for multiple luns*/
				command_status = SCSI_COMMAND_PASSED;
				g_tx_buf_ptr = (uint8_t)0;
				g_tx_buf_len = 0u;
                g_current_command_csw.data_residue = g_tx_buf_len;
                break;

            default:
				command_status = SCSI_COMMAND_FAILED;
				MSS_USB_MSD_prepare_sense_data(cbw->lun,
                                               SC_ILLEGAL_REQUEST,
                                               ASC_INVALID_CDB );

                g_current_command_csw.data_residue = g_bot_cbw.xfr_length;
                break;
        }
    }
    return command_status;
}

/***************************************************************************//**
 MSS_USB_MSD_prepare_sense_data() function prepares sense information required
 by the host in case there was an error in processing CBW. This information is
 sent as response to REQUEST_SENSE command.
 */
void
MSS_USB_MSD_prepare_sense_data
(
    uint8_t lun,
    uint8_t skey,
    uint8_t asc
)
{
	g_sense_response.sense_key = skey;
	g_sense_response.asc = asc;
}

/***************************************************************************//**
 MSS_USB_MSD_abort_cbw() function aborts the current command CBW.and prepares
 the driver to receive new CBW from HOST.
 */
void
MSS_USB_MSD_abort_cbw
(
    uint8_t lun
)
{
	/*Stall Bulk endpoints, Stall will be cleared by clear feature command.*/
	MSS_USB_core_rx_ep_stall( MSS_USB_RX_EP_1 );
	MSS_USB_core_tx_ep_stall( MSS_USB_TX_EP_1 );
	MSS_USB_core_rx_ep_flush_fifo( MSS_USB_RX_EP_1 );
	MSS_USB_core_tx_ep_flush_fifo( MSS_USB_TX_EP_1 );

    G_BOT_CBW_INIT();
	g_current_command_csw.data_residue = 0u;
	g_tx_buf_ptr = (uint8_t*)0;
	g_tx_buf_len = 0u;
	g_rx_buf_ptr = (uint8_t*)0;
	g_rx_buf_len = 0;

	/*Prepare to receive next command*/
	MSS_USB_device_rx_ep_read_prepare( MSS_USB_RX_EP_1, (uint8_t*)&g_bot_cbw, 31u);
}

/***************************************************************************//**
 MSS_USB_MSD_send_data() function send data to the host (DATA phase of IN transactions)
 in response to current CBW command.
 */
void
MSS_USB_MSD_send_data
(
    uint8_t* buf,
    uint32_t len
)
{
    /*TODO: check if the length can be accomodated in EP buffer*/
	g_bot_state = BOT_DATA_IN;
    MSS_USB_device_tx_ep_write( MSS_USB_TX_EP_1, buf, len );
	g_current_command_csw.data_residue -= len;
}

/***************************************************************************//**
 MSS_USB_MSD_receive_data() function prepares the driver to recieve data
 from USB host (DATA phase of OUT transactions) in response to the current CBW
 command.
 */
void
MSS_USB_MSD_receive_data
(
    uint8_t* buf,
    uint32_t len
)
{
	g_bot_state = BOT_DATA_OUT;
	MSS_USB_device_rx_ep_read_prepare( MSS_USB_RX_EP_1, g_rx_buf_ptr, len );
}

/***************************************************************************//**
 MSS_USB_MSD_send_csw() function sends status information to USB host 
 (DATA phase of IN transactions), indicating the Success/Failure status of the
 current CBW command.
 */
void
MSS_USB_MSD_send_csw
(
    mss_usb_msd_csw_status_t csw_status
)
{
    g_bot_state = BOT_SEND_STATUS;
	g_current_command_csw.status = csw_status;

    MSS_USB_device_tx_ep_write(MSS_USB_TX_EP_1,
                               (uint8_t*)&g_current_command_csw,
                               USB_MSD_BOT_CSW_LENGTH );

    MSS_USB_device_rx_ep_read_prepare( MSS_USB_RX_EP_1, (uint8_t*)&g_bot_cbw, 31 );
	g_current_command_csw.data_residue = 0u;
}

/***************************************************************************//**
 MSS_USB_MSD_process_read_10() function processes read command and calls application
 call-back function to read data from media.
 */
mss_usb_msd_csw_status_t
MSS_USB_MSD_process_read_10
(
    void
)
{
	uint32_t actual_read_len = 0u;

    uint32_t lba;
	uint32_t lba_addr;
	uint8_t* buf = 0;

	mss_usb_msd_csw_status_t command_status = SCSI_COMMAND_FAILED;

	/*Lba address is in command block byte 2 to 4 with MSD first.*/
    lba =  (g_bot_cbw.cmd_block[2] << 24u) |
           (g_bot_cbw.cmd_block[3] << 16u) |
           (g_bot_cbw.cmd_block[4] << 8u) |
            g_bot_cbw.cmd_block[5];

	/*Block size was received in USB_MSD_SCSI_READ_CAPACITY_10 command*/
	lba_addr = (lba * lun_capacity[g_bot_cbw.lun].blk_sz_len);

	/*
    Number of LBAs to be read is provided in command block byte 7 to 8,
    calculate the total size to be read
    */
	g_current_command_csw.data_residue = (( g_bot_cbw.cmd_block[7] << 8u ) | g_bot_cbw.cmd_block[8] ) *
                                            lun_capacity[g_bot_cbw.lun].blk_sz_len;

    actual_read_len = g_mss_usb_msd_media_ops->msd_media_read(g_bot_cbw.lun,
															&buf,
															lba_addr,
															g_current_command_csw.data_residue );

    if ( ( actual_read_len > g_current_command_csw.data_residue ) ||
         ( actual_read_len == 0u ) ||
         ( buf == 0u ) )
    {
        command_status = SCSI_COMMAND_FAILED;
    }
    else
    {
        g_tx_buf_ptr = buf;
        g_tx_buf_len = actual_read_len;

        /*
        Store the current LBA address so that it will be used to calculate new LBA
        address when remaining data is read from flash.
        */
        g_tx_current_lba_addr = lba_addr;
        command_status = SCSI_COMMAND_PASSED;
    }
    return (command_status);
}

/***************************************************************************//**
 MSS_USB_MSD_process_write_10() function processes Write command and calls application
 call-back function to acquire a buffer where the data will be receive.
 */
mss_usb_msd_csw_status_t
MSS_USB_MSD_process_write_10
(
    void
)
{
 	uint32_t lba;
	uint32_t lba_addr;
	uint8_t* write_buf;
	uint32_t app_write_len;
	mss_usb_msd_csw_status_t command_status = SCSI_COMMAND_FAILED;

	/*Lba address is in command block byte 2 to 4 with MSD first.*/
    lba = (g_bot_cbw.cmd_block[2] << 24u) |
          (g_bot_cbw.cmd_block[3] << 16u) |
          (g_bot_cbw.cmd_block[4] << 8u) |
           g_bot_cbw.cmd_block[5];

	/*Block size was received in USB_MSD_SCSI_READ_CAPACITY_10 command*/
	lba_addr = ( lba * lun_capacity[g_bot_cbw.lun].blk_sz_len );

	/*
    Number of LBA to be written is provided in command block byte 7 to 8,
    calculate the total size to be writen
    */
	g_current_command_csw.data_residue = ( ( g_bot_cbw.cmd_block[7] << 8) | g_bot_cbw.cmd_block[8]) *
                                            lun_capacity[g_bot_cbw.lun].blk_sz_len;



    write_buf = g_mss_usb_msd_media_ops->msd_media_acquire_write_buf( g_bot_cbw.lun,
                                                                    lba_addr,
                                                                    &app_write_len );

    if( ( write_buf != 0) && ( app_write_len != 0 ))
    {
        /*
        Since App is not capable of handling all data at the moment.
        Store the current LBA address so that it will be used to calculate new LBA
        address when remaining data is received.
        */
        g_rx_current_lba_addr = lba_addr;

        /*
        If application buffer is not big enough to handle all requested data from Host,
        then read the amount of data that the application can handle at the moment
        */
        if( app_write_len <= g_current_command_csw.data_residue )
        {
            g_rx_buf_len = app_write_len;
        }
        else
        {
            g_rx_buf_len = g_current_command_csw.data_residue;
        }

        g_rx_buf_ptr = write_buf;
        command_status = SCSI_COMMAND_PASSED;
    }
    else
    {
        command_status = SCSI_COMMAND_FAILED;
    }
	return(command_status);
}

#ifdef __cplusplus
}
#endif
