/*******************************************************************************
 * (c) Copyright 2012 Microsemi Corporation.  All rights reserved.
 *
 *  Microsemi Smartfusion2 USB device logical driver layer Implementation.	
 *  
 *
 *  This file implements HID class specific initialization and request handling.
 *
 * SVN $Revision: 5027 $
 * SVN $Date: 2013-02-01 19:06:39 +0530 (Fri, 01 Feb 2013) $
 */

#include "mss_usb_device_hid.h"
#include "mss_usb_device.h"

#ifdef __cplusplus
extern "C" {
#endif 

/***************************************************************************//**
* Local functions.
*/ 
uint8_t* MSS_USB_HID_get_descriptor_cb(uint8_t recepient, uint8_t type, uint32_t* length);
uint8_t MSS_USB_HID_init_cb(uint8_t cfgidx);
uint8_t MSS_USB_HID_deinit_cb(uint8_t cfgidx);
uint8_t MSS_USB_HID_process_request_cb(uint8_t** buf_pp, 
									uint8_t request,
									uint32_t* length,
									uint8_t* data_buf_p,
									uint32_t data_len);

uint8_t MSS_USB_HID_data_in_cb(mss_usb_ep_num_t num, uint8_t status);
uint8_t MSS_USB_HID_dataout_cb(mss_usb_ep_num_t num, uint8_t status, uint32_t rx_count);
uint8_t MSS_USB_HID_cep_datain_cb(uint8_t status);
uint8_t MSS_USB_HID_cep_dataout_cb(uint8_t status);

/* Call-back function structure definition needed by USB Device Core Driver. */
mss_usb_class_cb_t usb_hid_class_cb = {
    MSS_USB_HID_init_cb,
    MSS_USB_HID_deinit_cb,
	MSS_USB_HID_get_descriptor_cb,
	MSS_USB_HID_process_request_cb,
	MSS_USB_HID_data_in_cb,
    MSS_USB_HID_dataout_cb,
    MSS_USB_HID_cep_datain_cb,
    MSS_USB_HID_cep_dataout_cb
};

uint32_t g_tx_complete_status = 1u;

usb_hid_state_t g_usb_hid_state = USB_HID_NOT_CONFIGURED;

uint8_t hid_configuration_descriptor[FULL_CONFIG_DESCR_LENGTH] = {
    /*----------------------- Configuration Descriptor -----------------------*/
    CONFIGURATION_DESCR_LENGTH,         /* bLength */
    USB_CONFIGURATION_DESCRIPTOR_TYPE,  /* bDescriptorType */
    0x22u,                              /* wTotalLength LSB */
    0x00u,                              /* wTotalLength MSB */
    0x01u,                              /* bNumInterfaces */
    0x01u,                              /* bConfigurationValue */
    0x04u,                              /* iConfiguration */
    0xA0u,                              /* bmAttributes */
    0x32u,                              /* bMaxPower */
    /*------------------------- Interface Descriptor -------------------------*/
    INTERFACE_DESCR_LENGTH,             /* bLength */
    USB_INTERFACE_DESCRIPTOR_TYPE,       /* bDescriptorType */
    0x00u,                              /* bInterfaceNumber */
    0x00u,                              /* bAlternateSetting */
    0x01u,                              /* bNumEndpoints */
    0x03u,                              /* bInterfaceClass */
    0x01u,                              /* bInterfaceSubClass */
    0x02u,                              /* bInterfaceProtocol */
    0x00u,                              /* bInterface */
    /*---------------------------- HID Descriptor ----------------------------*/
    0x09u,                              /* bLength */
    0x21u,                              /* bDescriptorType */
    0x10u,                              /* bcdHID */
    0x01u,                      
    0x00u,                              /* bCountryCode */
    0x01u,                              /* bNumDescriptor */
    0x22u,                              /* bDescriptorType */
    0x34u,                              /* wDescriptorLength */
    0x00u,
    /*------------------------- Endpoint Descriptor --------------------------*/
    ENDPOINT_DESCR_LENGTH,              /* bLength */
    USB_ENDPOINT_DESCRIPTOR_TYPE,       /* bDescriptorType */
    0x81u,                              /* bEndpointAddress */
    0x03u,                              /* bmAttributes */
    0x08u,                              /* wMaxPacketSize LSB */
    0x00u,                              /* wMaxPacketSize MSB */
    0x0Au                               /* bInterval */
};

/*
Report descriptor: sent to the host during enumeration process.
This descriptors defines the format of the HID report.
*/
uint8_t report_descriptor[] = {
    0x05u, 0x01u,
    0x09u, 0x02u,
    0xA1u, 0x01u,
    0x09u, 0x01u,
    0xA1u, 0x00u,
    0x05u, 0x09u,
    0x19u, 0x01u,
    0x29u, 0x03u,
    0x15u, 0x00u,
    0x25u, 0x01u,
    0x75u, 0x01u,
    0x95u, 0x03u,
    0x81u, 0x02u,
    0x75u, 0x05u,
    0x95u, 0x01u,
    0x81u, 0x01u,
    0x05u, 0x01u,
    0x09u, 0x30u,
    0x09u, 0x31u,
    0x09u, 0x38u,
    0x15u, 0x81u,
    0x25u, 0x7Fu,
    0x75u, 0x08u,
    0x95u, 0x03u,
    0x81u, 0x06u,
    0xC0u, 0xC0u
};

uint8_t*
MSS_USB_HID_get_descriptor_cb
(
    uint8_t recepient,
    uint8_t type,
    uint32_t* length
)
{
	if( ( recepient == 0 ) && ( type ==2 ) )
    {
	   *length = sizeof( hid_configuration_descriptor );
        return( hid_configuration_descriptor );
	}
	if( ( recepient == 1 ) && ( type ==0x22 ) )
	{
	   *length = sizeof( report_descriptor );
		return( report_descriptor );
	}
	if( (recepient == 2) )
	{
	}
    return FAIL;    
}

uint8_t
MSS_USB_HID_init_cb
(
    uint8_t cfgidx
)
{
	MSS_USB_device_tx_ep_configure(MSS_USB_TX_EP_1,
							       0x200,
							       512,
							       512,
							       1,
							       DMA_DISABLE,
							       MSS_USB_DMA_CHANNEL1,
							       MSS_USB_XFR_INTERRUPT,
							       NO_ZLP_TO_XFR);
    g_usb_hid_state = USB_HID_CONFIGURED;
	return 1;
}

uint8_t
MSS_USB_HID_deinit_cb
(
    uint8_t cfgidx
)
{
    g_usb_hid_state = USB_HID_NOT_CONFIGURED;
    return 1;
}

uint8_t
MSS_USB_HID_process_request_cb
(	
    uint8_t** buf_pp, 
	uint8_t request,
	uint32_t* length,
	uint8_t* data_buf_p,
	uint32_t data_len	
)
{
    return SUCCESS;
}  
  
uint8_t
MSS_USB_HID_data_in_cb
(
    mss_usb_ep_num_t num,
    uint8_t status
)
{
	if( status > 1u )
	{
		MSS_USB_core_tx_ep_flush_fifo( MSS_USB_TX_EP_1 );
	}
	else
	{
		g_tx_complete_status = 1u;
	}
    return 1;
}

uint8_t
MSS_USB_HID_dataout_cb
(
	mss_usb_ep_num_t num,
	uint8_t status,
	uint32_t rx_count
)
{
    return SUCCESS;
}

uint8_t
MSS_USB_HID_cep_datain_cb
(
	uint8_t status
)
{
    return SUCCESS;    
}
uint8_t
MSS_USB_HID_cep_dataout_cb
(
    uint8_t status
)
{
    return SUCCESS;    
}

/***************************************************************************//**
* Exported function from HID Class driver.
*******************************************************************************/

/***************************************************************************//**
* See mss_usb_device_hid.h for details of how to use this function.
*/ 
uint32_t
MSS_USB_HID_tx_report
(
    uint8_t * buf,
    uint32_t length
)
{
    if(g_usb_hid_state == USB_HID_CONFIGURED)
    {
        g_tx_complete_status = 0;
        MSS_USB_device_tx_ep_write( MSS_USB_TX_EP_1, buf, length );
        return (1u);
    }
    else
    {
        return (0u);
    }
}

/***************************************************************************//**
* See mss_usb_device_hid.h for details of how to use this function.
*/ 
uint32_t
MSS_USB_HID_tx_done
(
    void
)
{
    return(g_tx_complete_status);
}

/***************************************************************************//**
* See mss_usb_device_hid.h for details of how to use this function.
*/ 
void
MSS_USB_HID_init
(
    void
)
{
    MSS_USB_device_set_class_cb_handler(&usb_hid_class_cb);
}

#ifdef __cplusplus
}
#endif
